module Main where

import Control.Parallel
import Control.Parallel.Strategies
import Control.DeepSeq
import Criterion.Main
import System.Random

qsort' [] = []
qsort' [x] = [x]
qsort' (x:xs) = qsort' [y | y <- xs, y<x]
  ++ [x]
  ++ qsort' [y | y <- xs, y>=x]

{- SkewHeap has been stolen from some stackoverflow thread -}

data SkewHeap a = Empty | SkewNode a (SkewHeap a) (SkewHeap a) deriving (Show)

(+++) :: Ord a => SkewHeap a -> SkewHeap a -> SkewHeap a
heap1@(SkewNode x1 l1 r1) +++ heap2@(SkewNode x2 l2 r2) 
  | x1 <= x2    = SkewNode x1 (heap2 +++ r1) l1 
  | otherwise = SkewNode x2 (heap1 +++ r2) l2
Empty +++ heap = heap
heap +++ Empty = heap

singleton x = SkewNode x Empty Empty

extractMin Empty = Nothing
extractMin (SkewNode x l r ) = Just (x , l +++ r )

newtype HeadList a = HeadList [a]

instance Eq a => Eq (HeadList a) where
  HeadList (a:_) == HeadList (b:_) = a == b

instance Ord a => Ord (HeadList a) where
  HeadList (a:_) <= HeadList (b:_) = a <= b
  HeadList (a:_) >= HeadList (b:_) = a >= b
  HeadList (a:_) <  HeadList (b:_) = a <  b
  HeadList (a:_) >  HeadList (b:_) = a >  b
  compare (HeadList (a:_)) (HeadList (b:_)) = compare a b

nestsort l = nestsort' [2000] l

nestsort' []     l = qsort' l
nestsort' (c:cs) l =
  merge $ foldr (+++) Empty $ map (singleton . HeadList) $
  withStrategy (parList (evalList rseq)) $
  map (nestsort' cs) $ chunk c l
  where
  merge s = case extractMin s of
    Nothing -> []
    Just (HeadList (x:xs), s') ->
      x : case xs of
            [] -> merge s'
            _  -> merge (s' +++ singleton (HeadList xs))
  chunk _ [] = []
  chunk n l  = l1 : chunk n l2
    where
    (l1, l2) = splitAt n l


main = defaultMain
   [--bench "qsort" (nf qsort randomInts),
    bench "nestsort" (nf nestsort randomInts)]

randomInts =
  take 200000 (randoms (mkStdGen 211570155)) :: [Integer]

